@@ -64,24 +64,34 @@ def safe_extract_all(star: tarfile.TarFile, topdir: pathlib.Path) -> None:
     """Validate the member names in the archive and extract them all."""
     members: Final = star.getmembers()
     paths: Final = [pathlib.Path(member.name) for member in members]
-    if not paths:
-        sys.exit("Expected at least one member in the source archive")
-    if paths[0].is_absolute():
-        sys.exit(f"Did not expect an absolute path {paths[0]} in the source archive")
+    match paths:
+        case []:
+            sys.exit("Expected at least one member in the source archive")
 
-    base_path: Final = paths[0].parts[0]
-    bad_paths: Final = [path for path in paths if path.parts[0] != base_path]
-    if bad_paths:
-        sys.exit(
-            f"Bad paths in the source archive, expected all of them to "
-            f"start with {base_path}: {bad_paths}",
-        )
+        case [first, *_] if first.is_absolute():
+            sys.exit(f"Did not expect an absolute path {first} in the source archive")
 
-    bad_dirs: Final = [path for path in paths if ".." in path.parts]
-    if bad_dirs:
-        sys.exit(f"Bad paths in the source archive, none of them should contain '..': {bad_paths}")
+        case [first, *_]:
+            base_path: Final = first.parts[0]
+            bad_paths: Final = [path for path in paths if path.parts[0] != base_path]
+            if bad_paths:
+                sys.exit(
+                    f"Bad paths in the source archive, expected all of them to "
+                    f"start with {base_path}: {bad_paths}",
+                )
 
-    star.extractall(topdir, members=members)  # noqa: S202
+            bad_dirs: Final = [path for path in paths if ".." in path.parts]
+            if bad_dirs:
+                sys.exit(
+                    f"Bad paths in the source archive, "
+                    f"none of them should contain '..': {bad_paths}",
+                )
+
+            if sys.version_info >= (3, 12):
+                # We *sincerely* hope our own sdist tarball does not contain anything weird
+                star.extractall(topdir, members=members, filter="data")
+            else:
+                star.extractall(topdir, members=members)  # noqa: S202
 
 
 def extract_sdist(sdist: pathlib.Path, tempd: pathlib.Path) -> pathlib.Path:
@@ -94,25 +104,33 @@ def extract_sdist(sdist: pathlib.Path, tempd: pathlib.Path) -> pathlib.Path:
 
     with tarfile.open(sdist, mode="r") as star:
         safe_extract_all(star, topdir)
-        entries: Final = sorted(path for path in topdir.iterdir())
-        if len(entries) != 1:
-            sys.exit(f"Expected {sdist} to contain a single directory, got {entries!r}")
+        match sorted(path for path in topdir.iterdir()):
+            case [testdir] if not testdir.is_dir() or not testdir.name.startswith(
+                ("test_stages-", "test-stages-"),
+            ):
+                sys.exit(
+                    f"Expected {sdist} to contain a single `test-stages-*` directory, "
+                    f"got {testdir}",
+                )
 
-    testdir: Final = entries[0]
-    if not testdir.is_dir() or not testdir.name.startswith(("test_stages-", "test-stages-")):
-        sys.exit(f"Expected {sdist} to contain a single `test-stages-*` directory, got {testdir}")
-    return testdir
+            case [testdir]:
+                return testdir
+
+            case entries:
+                sys.exit(f"Expected {sdist} to contain a single directory, got {entries!r}")
 
 
 def adapt_pyproject(testdir: pathlib.Path) -> None:
     """Disable this selftest to avoid infinite recursion."""
     projfile: Final = testdir / "pyproject.toml"
     projdata: Final = tomllib.loads(projfile.read_text(encoding="UTF-8"))
+    test_stages: Final[dict[str, list[str]]] = projdata["tool"]["test-stages"]
+    match test_stages["stages"]:
+        case [*others, str(last)] if last.startswith("@tests"):
+            test_stages["stages"] = [*others, f"{last} and not selftest"]
 
-    stages: Final = projdata["tool"]["test-stages"]["stages"]
-    if not stages[-1].startswith("@tests"):
-        sys.exit(f"Expected a `@tests...` test-stages entry, got {stages!r}")
-    stages[-1] = f"{stages[-1]} and not selftest"
+        case stages:
+            sys.exit(f"Expected a `@tests...` test-stages entry, got {stages!r}")
 
     projfile.write_text(tomli_w.dumps(projdata), encoding="UTF-8")
 
@@ -171,7 +189,7 @@ def run_tox(testdir: pathlib.Path) -> None:
         encoding="UTF-8",
         env=utf8_env,
     ):
-        sys.exit("A run without any -p option output {blurb!r}")
+        sys.exit(f"A run without any -p option output {blurb!r}")
 
     if blurb in subprocess.check_output(
         ["tox-stages", "run", "@selftest", "-p", "1"],
@@ -179,7 +197,7 @@ def run_tox(testdir: pathlib.Path) -> None:
         encoding="UTF-8",
         env=utf8_env,
     ):
-        sys.exit("A `-p 1` run did not output {blurb!r}")
+        sys.exit(f"A `-p 1` run did not output {blurb!r}")
 
     if blurb not in subprocess.check_output(
         ["tox-stages", "run", "@selftest", "-p", "7"],
@@ -187,7 +205,7 @@ def run_tox(testdir: pathlib.Path) -> None:
         encoding="UTF-8",
         env=utf8_env,
     ):
-        sys.exit("A `-p 7` run output {blurb!r}")
+        sys.exit(f"A `-p 7` run output {blurb!r}")
 
     subprocess.check_call(["tox-stages", "run"], cwd=testdir, env=env)
 
